@font-face {
  font-family: 'fontello';
  src: url('../font/fontello.eot?13439518');
  src: url('../font/fontello.eot?13439518#iefix') format('embedded-opentype'),
       url('../font/fontello.woff?13439518') format('woff'),
       url('../font/fontello.ttf?13439518') format('truetype'),
       url('../font/fontello.svg?13439518#fontello') format('svg');
  font-weight: normal;
  font-style: normal;
}
/* Chrome hack: SVG is rendered more smooth in Windozze. 100% magic, uncomment if you need it. */
/* Note, that will break hinting! In other OS-es font will be not as sharp as it could be */
/*
@media screen and (-webkit-min-device-pixel-ratio:0) {
  @font-face {
    font-family: 'fontello';
    src: url('../font/fontello.svg?13439518#fontello') format('svg');
  }
}
*/

 [class^="icon-"]:before, [class*=" icon-"]:before {
  font-family: "fontello";
  font-style: normal;
  font-weight: normal;
  speak: none;

  display: inline-block;
  text-decoration: inherit;
  width: 1em;
  margin-right: .2em;
  text-align: center;
  /* opacity: .8; */

  /* For safety - reset parent styles, that can break glyph codes*/
  font-variant: normal;
  text-transform: none;

  /* fix buttons height, for twitter bootstrap */
  line-height: 1em;

  /* Animation center compensation - margins should be symmetric */
  /* remove if not needed */
  margin-left: .2em;

  /* you can be more comfortable with increased icons size */
  /* font-size: 120%; */

  /* Uncomment for 3D effect */
  /* text-shadow: 1px 1px 1px rgba(127, 127, 127, 0.3); */
}

.icon-mail-1:before { content: '\e807'; } /* '' */
.icon-up-circle:before { content: '\e80c'; } /* '' */
.icon-down-circle-1:before { content: '\e80d'; } /* '' */
.icon-left-circle-1:before { content: '\e80e'; } /* '' */
.icon-right-circle-1:before { content: '\e80f'; } /* '' */
.icon-up-circle-1:before { content: '\e810'; } /* '' */
.icon-mail-2:before { content: '\e805'; } /* '' */
.icon-mail:before { content: '\e806'; } /* '' */
.icon-plus:before { content: '\e808'; } /* '' */
.icon-left-open:before { content: '\e801'; } /* '' */
.icon-right-open:before { content: '\e802'; } /* '' */
.icon-up-open:before { content: '\e803'; } /* '' */
.icon-down-circle:before { content: '\e809'; } /* '' */
.icon-left-circle:before { content: '\e80a'; } /* '' */
.icon-right-circle:before { content: '\e80b'; } /* '' */
.icon-down-open:before { content: '\e800'; } /* '' */
